@@ -16,7 +16,7 @@ import dask
 from dask.system import CPU_COUNT
 
 from distributed import Nanny
-from distributed.cli.utils import check_python_3, install_signal_handlers
+from distributed.cli.utils import install_signal_handlers
 from distributed.comm import get_address_host_port
 from distributed.deploy.utils import nprocesses_nthreads
 from distributed.preloading import validate_preload_argv
@@ -56,7 +56,7 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
     "--worker-port",
     default=None,
     help="Serving computation port, defaults to random. "
-    "When creating multiple workers with --nprocs, a sequential range of "
+    "When creating multiple workers with --nworkers, a sequential range of "
     "worker ports may be used by specifying the first and last available "
     "ports like <first-port>:<last-port>. For example, --worker-port=3000:3026 "
     "will use ports 3000, 3001, ..., 3025, 3026.",
@@ -65,7 +65,7 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
     "--nanny-port",
     default=None,
     help="Serving nanny port, defaults to random. "
-    "When creating multiple nannies with --nprocs, a sequential range of "
+    "When creating multiple nannies with --nworkers, a sequential range of "
     "nanny ports may be used by specifying the first and last available "
     "ports like <first-port>:<last-port>. For example, --nanny-port=3000:3026 "
     "will use ports 3000, 3001, ..., 3025, 3026.",
@@ -127,18 +127,28 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
 @click.option(
     "--nprocs",
     type=str,
-    default=1,
+    default=None,
+    show_default=True,
+    help="Deprecated. Use '--nworkers' instead. Number of worker processes to "
+    "launch. If negative, then (CPU_COUNT + 1 + nprocs) is used. "
+    "Set to 'auto' to set nprocs and nthreads dynamically based on CPU_COUNT",
+)
+@click.option(
+    "--nworkers",
+    "n_workers",  # This sets the Python argument name
+    type=str,
+    default=None,
     show_default=True,
     help="Number of worker processes to launch. "
-    "If negative, then (CPU_COUNT + 1 + nprocs) is used. "
-    "Set to 'auto' to set nprocs and nthreads dynamically based on CPU_COUNT",
+    "If negative, then (CPU_COUNT + 1 + nworkers) is used. "
+    "Set to 'auto' to set nworkers and nthreads dynamically based on CPU_COUNT",
 )
 @click.option(
     "--name",
     type=str,
     default=None,
     help="A unique name for this worker like 'worker-1'. "
-    "If used with --nprocs then the process number "
+    "If used with --nworkers then the process number "
     "will be appended like name-0, name-1, name-2, ...",
 )
 @click.option(
@@ -174,7 +184,7 @@ pem_file_option_type = click.Path(exists=True, resolve_path=True)
     default=None,
     help='Resources for task constraints like "GPU=2 MEM=10e9". '
     "Resources are applied separately to each worker process "
-    "(only relevant when starting multiple worker processes with '--nprocs').",
+    "(only relevant when starting multiple worker processes with '--nworkers').",
 )
 @click.option(
     "--scheduler-file",
@@ -250,6 +260,7 @@ def main(
     nanny_port,
     nthreads,
     nprocs,
+    n_workers,
     nanny,
     name,
     pid_file,
@@ -273,7 +284,7 @@ def main(
     enable_proctitle_on_current()
     enable_proctitle_on_children()
 
-    if bokeh_port is not None:
+    if bokeh_port is not None:  # pragma: no cover
         warnings.warn(
             "The --bokeh-port flag has been renamed to --dashboard-address. "
             "Consider adding ``--dashboard-address :%d`` " % bokeh_port
@@ -295,23 +306,38 @@ def main(
         if v is not None
     }
 
-    if nprocs == "auto":
-        nprocs, nthreads = nprocesses_nthreads()
-    else:
-        nprocs = int(nprocs)
-
-    if nprocs < 0:
-        nprocs = CPU_COUNT + 1 + nprocs
-
-    if nprocs <= 0:
+    if nprocs is not None and n_workers is not None:
         logger.error(
-            "Failed to launch worker. Must specify --nprocs so that there's at least one process."
+            "Both --nprocs and --nworkers were specified. Use --nworkers only."
+        )
+        sys.exit(1)
+    elif nprocs is not None:
+        warnings.warn(
+            "The --nprocs flag will be removed in a future release. It has been "
+            "renamed to --nworkers.",
+            FutureWarning,
+        )
+        n_workers = nprocs
+
+    if n_workers == "auto":
+        n_workers, nthreads = nprocesses_nthreads()
+    elif n_workers is None:
+        n_workers = 1
+    else:
+        n_workers = int(n_workers)
+
+    if n_workers < 0:
+        n_workers = CPU_COUNT + 1 + n_workers
+
+    if n_workers <= 0:
+        logger.error(
+            "Failed to launch worker. Must specify --nworkers so that there's at least one process."
         )
         sys.exit(1)
 
-    if nprocs > 1 and not nanny:
+    if n_workers > 1 and not nanny:
         logger.error(
-            "Failed to launch worker.  You cannot use the --no-nanny argument when nprocs > 1."
+            "Failed to launch worker.  You cannot use the --no-nanny argument when n_workers > 1."
         )
         sys.exit(1)
 
@@ -322,10 +348,10 @@ def main(
         )
         sys.exit(1)
 
-    if nprocs > 1 and listen_address:
+    if n_workers > 1 and listen_address:
         logger.error(
             "Failed to launch worker. "
-            "You cannot specify --listen-address when nprocs > 1."
+            "You cannot specify --listen-address when n_workers > 1."
         )
         sys.exit(1)
 
@@ -349,7 +375,7 @@ def main(
         else:
             # if contact address is not present we use the listen_address for contact
             contact_address = listen_address
-    except ValueError as e:
+    except ValueError as e:  # pragma: no cover
         logger.error("Failed to launch worker. " + str(e))
         sys.exit(1)
 
@@ -359,7 +385,7 @@ def main(
         port = worker_port
 
     if not nthreads:
-        nthreads = CPU_COUNT // nprocs
+        nthreads = CPU_COUNT // n_workers
 
     if pid_file:
         with open(pid_file, "w") as f:
@@ -420,17 +446,17 @@ def main(
             dashboard=dashboard,
             dashboard_address=dashboard_address,
             name=name
-            if nprocs == 1 or name is None or name == ""
+            if n_workers == 1 or name is None or name == ""
             else str(name) + "-" + str(i),
             **kwargs,
         )
-        for i in range(nprocs)
+        for i in range(n_workers)
     ]
 
     async def close_all():
         # Unregister all workers from scheduler
         if nanny:
-            await asyncio.gather(*[n.close(timeout=2) for n in nannies])
+            await asyncio.gather(*(n.close(timeout=2) for n in nannies))
 
     signal_fired = False
 
@@ -443,7 +469,7 @@ def main(
 
     async def run():
         await asyncio.gather(*nannies)
-        await asyncio.gather(*[n.finished() for n in nannies])
+        await asyncio.gather(*(n.finished() for n in nannies))
 
     install_signal_handlers(loop, cleanup=on_signal)
 
@@ -454,16 +480,11 @@ def main(
         if not signal_fired:
             logger.info("Timed out starting worker")
         sys.exit(1)
-    except KeyboardInterrupt:
+    except KeyboardInterrupt:  # pragma: no cover
         pass
     finally:
         logger.info("End worker")
 
 
-def go():
-    check_python_3()
-    main()
-
-
 if __name__ == "__main__":
-    go()
+    main()  # pragma: no cover
